from MLDashboard.DashboardModules.Module import Module
from MLDashboard.MLCommunicationBackend import Message
from PIL import Image
from typing import List

def compareImages(imga, imgb):
    if imga == imgb:
        return True

    if list(imga.getdata()) == list(imgb.getdata()):
        return True
    return False

def compareData(itemsa, itemsb, tol):
    for i in range(0, len(itemsa)):
        if abs(itemsa[i] - itemsb[i]) > tol:
            return False
    return True

class ImageModule(Module):
    """Base class for modules that rely on rendering images"""
    def __init__(self, ax, config, title, datarequesttype, reqkeys: List[str] = None):
        basereqkeys = ["width", "height", "rows", "cols", "refreshrate"]
        if reqkeys is not None:
            basereqkeys += reqkeys
        super().__init__(ax, config, title, noticks=True, reqkeys=basereqkeys)

        #defaulting
        if 'conversion' not in self.config:
            self.config['conversion'] = 'L' #grayscale
        if 'cmap' not in self.config:
            self.config['cmap'] = 'gray'

        #storage to quickly swap images
        self.axes = []
        self.axtables = []
        self.text = []
        self.imgs = []
        self.isclear = []

        self.refreshtimer = 0
        self.initreq = False
        self.imgcounter = 0 #index to grab images from
        self.datarequesttype = datarequesttype

        #old ax info
        self.axx = 0
        self.axy = 0
        self.axw = 0
        self.axh = 0

    def shouldRequest(self):
        if not self.initreq:
            self.initreq = True
            return True

        if self.config["refreshrate"] == "once":
            if self.initreq:
                return False
            self.initreq = True
            return True

        if self.config["refreshrate"] <= self.refreshtimer:
            self.refreshtimer = 0
            return True

        self.refreshtimer += 1
        return False

    def generateRequest(self, attempts=None):
        if self.shouldRequest():
            num = self.config['rows'] * self.config['cols']
            d = {"num": num, "startingindex": self.imgcounter, "attempts": attempts}
            self.imgcounter += num
            return [Message(self.datarequesttype, d)]


    def createImages(self, rawdata):
        images = []
        for image in rawdata:
            image = image.reshape(self.config['width'],self.config['height'])
            img = Image.fromarray(image * 255)
            images.append(img.convert(self.config['conversion']))
        return images

    def updateImageGrid(self, imgs = None, text = None, color=None):
        """Will not rerender unless it is necessary."""
        if imgs is None:
            imgs = self.imgs
        if text is None:
            text = self.text
        if color is None:
            color = ['black'] * len(text)

        if len(imgs) > 0:
            b = self.ax.get_position()
            x = b.x0
            y = b.y0
            width = b.width
            height = b.height

            rerender = not compareData([x, y, width, height], [self.axx, self.axy, self.axw, self.axh], 0.01)
            self.axx = x
            self.axy = y
            self.axw = width
            self.axh = height

            rows = self.config['rows']
            cols = self.config['cols']

            imgwidth = (width / cols)
            imgheight = (height / rows)

            counter = 0
            ypos = y
            for row in range(0, rows):
                xpos = x
                for col in range(0, cols):
                    xcoord = xpos + imgwidth / 6
                    ycoord = ypos + imgheight / 8
                    truewidth = imgwidth / 1.5
                    trueheight = imgheight / 1.5

                    if counter >= len(self.axes):
                        if counter < len(imgs):
                            ax, table = self.displayImage(xcoord, ycoord, truewidth, trueheight,
                                                               imgs[counter], text[counter], color[counter])
                            self.axes.append(ax)
                            self.axtables.append(table)
                            self.isclear.append(False)
                    else:
                        if rerender:
                            self.axes[counter].set_position((xcoord, ycoord, truewidth, trueheight))

                        if counter < len(imgs):
                            if counter >= len(self.text) or text[counter] != self.text[counter]:
                                t = self.axtables[counter][0,0].get_text()
                                t.set_text(text[counter])
                                t.set_color(color[counter])
                                t.set_fontsize(12) #this gets auto scaled down
                            if counter >= len(self.imgs) or not compareImages(imgs[counter], self.imgs[counter]):
                                self.axes[counter].imshow(imgs[counter], cmap=self.config['cmap'])
                                self.isclear[counter] = False
                                self.axes[counter].axis('on')
                        else:
                            if not self.isclear[counter]:
                                self.axes[counter].clear()
                                self.isclear[counter] = True
                                self.axes[counter].axis('off')

                    counter += 1
                    xpos += imgwidth
                ypos += imgheight

        self.imgs = imgs
        self.text = text


    def displayImage(self, x, y, width, height, img, text, textcolor='black'):
        #imshow will only shrink and resize axes
        fig = self.ax.get_figure()
        ax = fig.add_axes((x, y, width, height))
        ax.imshow(img, cmap=self.config['cmap'])
        ax.tick_params(axis='both', which='both', bottom=False, top=False,
                                labelbottom=False, right=False, left=False, labelleft=False)

        table = ax.table([[text]], loc='top', colWidths=[1.4], cellLoc='center')
        table[0,0].get_text().set_color(textcolor)
        table[0, 0].set_height(0.2)
        table[0,0].get_text().set_fontsize(12)  # this gets auto scaled down
        table[0,0].set_linewidth(0)
        return ax, table


import warnings

class MessageMode:
    #control characters
    Start = 0
    End = 1
    ForceUpdate = 2
    CustomData = 3 #handles status module comm
    Command = 4 #handles data return

    #callback messages
    Train_Begin = 10
    Train_End = 11
    Predict_Begin = 12
    Predict_End = 13
    Test_Begin = 14 #evaluate
    Test_End = 15

    Epoch_Begin = 16 #training only
    Epoch_End = 17

    Train_Batch_Begin = 18 #batch methods
    Train_Batch_End = 19
    Test_Batch_Begin = 20
    Test_Batch_End = 21
    Predict_Batch_Begin = 22
    Predict_Batch_End = 23

    #data request messages
    Train_Set_Sample = 30
    Test_Set_Sample = 31
    Pred_Sample = 32 #from test set
    Pred_Sample_Train = 33
    Wrong_Pred_Sample = 34 #from test set
    Wrong_Pred_Sample_Train = 35

def getMode(x: int):
    for y in [a for a in dir(MessageMode) if not a.startswith('__')]:
        if MessageMode().__getattribute__(y) == x:
            return y
    warnings.warn("No valid mode found.")

class Message:
    def __init__(self, mode: int, body: dict):
        self.mode = mode
        self.body = body

    def __repr__(self):
        return ("Message with mode: " + getMode(self.mode) + " and data payload: " + str(self.body) +
                " At location: " + object.__repr__(self))



from MLDashboard.MLCommunicationBackend import Message, MessageMode
from tensorflow.keras.callbacks import Callback
import numpy as np
import warnings
from typing import List


class CallbackConfig:
    """Customize when data is send to the dashboard"""
    def __init__(self, send_on_train_start = True, send_on_train_end = True, send_on_test_start = True,
                 send_on_test_end = True, send_on_predict_start = True, send_on_predict_end = True,
                 send_on_epoch_end = True, send_on_batch_end = False, send_on_test_batch_end = False,
                 send_on_predict_batch_end = False, force_update_on_epoch_end=True):

        self.send_on_train_start = send_on_train_start
        self.send_on_train_end = send_on_train_end
        self.send_on_test_start = send_on_test_start
        self.send_on_test_end = send_on_test_end
        self.send_on_predict_start = send_on_predict_start
        self.send_on_predict_end = send_on_predict_end

        self.send_on_epoch_end = send_on_epoch_end
        self.send_on_batch_end = send_on_batch_end #this can quickly overwhelm the dashboard
        self.send_on_test_batch_end = send_on_test_batch_end
        self.send_on_predict_batch_end = send_on_predict_batch_end

        #batch begin does not contain any keys

        self.force_update_on_epoch_end = force_update_on_epoch_end #redraws the screen

#region Callbacks
class DashboardCallbacks(Callback):
    def __init__(self, updatelist: List[Message], returnlist: List[Message], model, x_train, y_train, x_test, y_test,
                 prediction_labels, config: CallbackConfig):
        """
        This inherits from Tensorflow callbacks and connects the model training to the dashboard. Can be customized
        by overriding the custom_on functions.

        :param updatelist: List from dashboard creation
        :param returnlist: List from dashboard creation
        :param model: Tensorflow model
        :param x_train: Training set features
        :param y_train: Training set output
        :param x_test: Test set features
        :param y_test: Test set output
        :param prediction_labels: Allows images to be labeled with friendly text
        :param config: Customize when data is sent
        """

        super().__init__()
        self.updatelist = updatelist
        self.returnlist = returnlist
        self.model = model
        self.x_train = x_train
        self.y_train = y_train
        self.x_test = x_test
        self.y_test = y_test

        self.predictionlabels = prediction_labels

        self.config = config

        self.handleDataRequest()

    def label(self, data):
        labels = []
        for item in data:
            labels.append(self.predictionlabels[item])
        return labels

    def predlabel(self, data):
        labels = []
        for item in data:
            labels.append(self.predictionlabels[np.argmax(item)])
        return labels

    def sample(self, data, x, y):
        start = data.body["startingindex"]
        num = data.body["num"] + start
        x = x[start:num]
        y = y[start:num]
        return {"x": x, "y": self.label(y)}

    def predsample(self, data, x, y):
        start = data.body["startingindex"]
        num = data.body["num"] + start
        pred = self.predlabel(self.model.predict(x[start:num]))
        x = x[start:num]
        y = y[start:num]
        return {'x': x, 'y': self.label(y), 'pred': pred}

    def wrongpredsample(self, data, x, y):
        maxnum = data.body["num"]
        attempts = data.body["attempts"]

        preds = self.predlabel(self.model.predict(x[0:attempts]))
        y = self.label(y)
        features = []
        wrongpreds = []
        correctresult = []
        for i in range(0, attempts):
            if preds[i] != y[i]:
                features.append(x[i])
                wrongpreds.append(preds[i])
                correctresult.append(y[i])

            if len(features) >= maxnum:
                break
        return {'x': features, 'y': correctresult, 'pred': wrongpreds}

    # on epoch begin
    def handleDataRequest(self):
        rmlist = []
        for index, item in enumerate(self.returnlist):
            if item.mode == MessageMode.Train_Set_Sample:
                rmlist.append(index)
                self.updatelist.append(Message(MessageMode.Train_Set_Sample,
                                               self.sample(item, self.x_train, self.y_train)))

            elif item.mode == MessageMode.Test_Set_Sample:
                rmlist.append(index)
                self.updatelist.append(Message(MessageMode.Test_Set_Sample,
                                               self.sample(item, self.x_test, self.y_test)))

            elif item.mode == MessageMode.Pred_Sample:
                rmlist.append(index)
                self.updatelist.append(Message(MessageMode.Pred_Sample,
                                               self.predsample(item, self.x_test, self.y_test)))


            elif item.mode == MessageMode.Pred_Sample_Train:
                rmlist.append(index)
                self.updatelist.append(Message(MessageMode.Pred_Sample_Train,
                                               self.predsample(item, self.x_train, self.y_train)))

            elif item.mode == MessageMode.Wrong_Pred_Sample:
                rmlist.append(index)
                self.updatelist.append(Message(MessageMode.Wrong_Pred_Sample,
                                               self.wrongpredsample(item, self.x_test, self.y_test)))

            elif item.mode == MessageMode.Wrong_Pred_Sample_Train:
                rmlist.append(index)
                self.updatelist.append(Message(MessageMode.Wrong_Pred_Sample_Train,
                                               self.wrongpredsample(item, self.x_train, self.y_train)))

        rmlist.reverse()
        for index in rmlist:
            self.returnlist.pop(index)

    # on epoch end
    def handleCommands(self, allowstop=True):
        rmlist: List[int] = []
        for index, item in enumerate(self.returnlist):
            if item.mode == MessageMode.Command:
                rmlist.append(index)
                if item.body['command'] == 'stop':
                    if allowstop:
                        print("Training manually stopped.")
                        self.model.stop_training = True
                    else:
                        warnings.warn("Stop was triggered but training has already exited.")
                elif item.body['command'] == 'save':
                    print("Model manually saved.")
                    self.model.save(input("File name: "))

        rmlist.reverse()
        for index in rmlist:
            self.returnlist.pop(index)

    def HandleRemaingCommands(self):  # TODO - handle remaining data requests
        """
        This should be called after the dashboard exits.
        """
        self.handleCommands(allowstop=False)
        if len(self.returnlist) > 0:
            warnings.warn("We couldn't handle these remaining requests: ")
        for req in self.returnlist:
            print(req)

    #region callbacks
    def on_train_begin(self, logs=None):
        if self.config.send_on_train_start:
            self.updatelist.append(Message(MessageMode.Train_Begin, logs))
        self.custom_on_train_begin(logs)

    def on_train_end(self, logs=None):
        if self.config.send_on_train_end:
            self.updatelist.append(Message(MessageMode.Train_End, logs))
        self.custom_on_train_end(logs)

    def on_test_begin(self, logs=None):
        if self.config.send_on_test_start:
            self.updatelist.append(Message(MessageMode.Test_Begin, logs))
        self.custom_on_test_begin(logs)

    def on_test_end(self, logs=None):
        if self.config.send_on_test_end:
            self.updatelist.append(Message(MessageMode.Test_End, logs))
        self.custom_on_test_end(logs)

    def on_predict_begin(self, logs=None):
        if self.config.send_on_predict_start:
            self.updatelist.append(Message(MessageMode.Predict_Begin, logs))
        self.custom_on_predict_begin(logs)

    def on_predict_end(self, logs=None):
        if self.config.send_on_predict_end:
            self.updatelist.append(Message(MessageMode.Predict_End, logs))
        self.custom_on_predict_end(logs)

    def on_epoch_begin(self, epoch, logs=None):
        logs['epoch'] = epoch
        self.handleDataRequest()
        self.custom_on_epoch_begin(logs)

    def on_epoch_end(self, epoch, logs=None):
        logs['epoch'] = epoch
        if self.config.send_on_epoch_end:
            self.updatelist.append(Message(MessageMode.Epoch_End, logs))
        if self.config.force_update_on_epoch_end:
            self.updatelist.append(Message(MessageMode.ForceUpdate, {}))
        self.handleCommands()
        self.custom_on_epoch_end(logs)

    def on_train_batch_begin(self, batch, logs=None):
        logs['batch'] = batch
        self.custom_on_train_batch_begin(logs)

    def on_test_batch_begin(self, batch, logs=None):
        logs['batch'] = batch
        self.custom_on_test_batch_begin(logs)

    def on_predict_batch_begin(self, batch, logs=None):
        logs['batch'] = batch
        self.custom_on_predict_batch_begin(logs)


    def on_train_batch_end(self, batch, logs=None):
        logs['batch'] = batch
        if self.config.send_on_batch_end:
            self.updatelist.append(Message(MessageMode.Train_Batch_End, logs))
        self.custom_on_train_batch_end(batch)

    def on_test_batch_end(self, batch, logs=None):
        logs['batch'] = batch
        if self.config.send_on_batch_end:
            self.updatelist.append(Message(MessageMode.Test_Batch_End, logs))
        self.custom_on_test_batch_end(batch)

    def on_predict_batch_end(self, batch, logs=None):
        logs['batch'] = batch
        if self.config.send_on_batch_end:
            self.updatelist.append(Message(MessageMode.Predict_Batch_End, logs))
        self.custom_on_predict_batch_end(batch)
    #endregion
    #region custom overrides
    def custom_on_train_begin(self, logs):
        pass

    def custom_on_train_end(self, logs):
        pass

    def custom_on_test_begin(self, logs):
        pass

    def custom_on_test_end(self, logs):
        pass

    def custom_on_predict_begin(self, logs):
        pass

    def custom_on_predict_end(self, logs):
        pass

    def custom_on_epoch_begin(self, logs):
        pass

    def custom_on_epoch_end(self, logs):
        pass

    def custom_on_train_batch_begin(self, logs):
        pass

    def custom_on_test_batch_begin(self, logs):
        pass

    def custom_on_predict_batch_begin(self, logs):
        pass

    def custom_on_train_batch_end(self, logs):
        pass

    def custom_on_test_batch_end(self, logs):
        pass

    def custom_on_predict_batch_end(self, logs):
        pass
    #endregion
#endregion Callbacks


from MLDashboard.MLCommunicationBackend import Message
from typing import List
import warnings

class Module:
    def __init__(self, ax, config: dict, title: str, noticks=False, reqkeys: List[str] = None):
        """
        Creates a module in a axes

        :param ax: matplotlib axes
        :param config: module specific config
        :param title: Placed above plot
        :param noticks: Should plot have tick marks
        :param reqkeys: List of required keys in config
        """
        self.ax = ax
        self.config = config
        self.ax.set_title(title)

        if noticks:
            self.ax.tick_params(axis='both', which='both', bottom=False, top=False,
                                labelbottom=False, right=False, left=False, labelleft=False)

        if reqkeys is not None:
            for key in reqkeys:
                if key not in config:
                    raise Exception("Config for module: " + str(title) + " missing key: " + str(key))

    def initialRequests(self):
        pass

    def update(self, data: Message):
        if self == self:
            pass
        warnings.warn("Update method should have been overriden, but it was not.")
        warnings.warn("Update method called with: " + str(data))


from MLDashboard.DashboardModules.Module import Module
from MLDashboard.MLCommunicationBackend import Message, MessageMode
import matplotlib.pyplot as pyplot
from matplotlib.widgets import Button
from mpl_toolkits.axes_grid1.inset_locator import InsetPosition
import warnings
import copy
from typing import List


warnings.filterwarnings(action='ignore',
                        message=('This figure includes Axes that are not compatible with tight_layout, '
                                 'so results might be incorrect.')) #stops buttons from being broken

def createButtonWithingAxes(ax, x, y, width, height, text):
    button_ax = pyplot.axes([0, 0, 1, 1])
    ip = InsetPosition(ax, [x, y, width, height])
    button_ax.set_axes_locator(ip)
    return Button(button_ax, text)

class ControlButtons(Module):
    def __init__(self, ax, config):
        """Contains buttons to stop training and save model"""
        super().__init__(ax, config, "Control Buttons", noticks=True)
        self.stopbutton = createButtonWithingAxes(self.ax, 0.2, 0.2, 0.2, 0.1, "Stop Training")
        self.savebutton = createButtonWithingAxes(self.ax, 0.5, 0.2, 0.2, 0.1, "Save Model")
        self.stopbutton.on_clicked(self.stopFunc)
        self.savebutton.on_clicked(self.saveFunc)

        self.internalreturnlist: List[Message] = []

    def update(self, data: Message):
        out = copy.deepcopy(self.internalreturnlist)
        self.internalreturnlist = []
        return out

    def stopFunc(self, event):
        if event == event:
            pass
        self.internalreturnlist.append(Message(MessageMode.Command, {'command': 'stop'}))

    def saveFunc(self, event):
        if event == event:
            pass
        self.internalreturnlist.append(Message(MessageMode.Command, {'command': 'save'}))


from MLDashboard.DashboardModules.Module import Module
from MLDashboard.MLCommunicationBackend import Message

class EmptyModule(Module):
    def __init__(self, ax, config):
        """Module with no functionality, used to make a grid of modules"""
        super().__init__(ax, config, "", noticks=True)
        self.ax.axis([0, 10, 0, 10])
        self.ax.text(5, 5, "This Page Intentionally Left Blank", ha='center', va='center')

    def update(self, data: Message):
        pass



from MLDashboard.DashboardModules.Module import Module
from MLDashboard.MLCommunicationBackend import Message, MessageMode
import warnings

class LossMetricsNumerical(Module):
    """Shows loss, accuracy, and other metrics by listing stats. Also shows evaluation stats at end."""
    def __init__(self, ax, config):
        super().__init__(ax, config, "Loss Metrics Info", noticks=True)
        self.ax.axis([0, 10, 0, 10])
        self.epochtext = self.ax.text(1, 9, "Epoch: ")
        self.losstext = self.ax.text(1, 8, "Loss: ")
        self.othertext = []

        self.pos = 0


    def update(self, data: Message):
        if data.mode == MessageMode.Epoch_End:
            self.epochtext.set_text("Epoch: " + str(round(data.body["epoch"] + 1, 3)))
            self.losstext.set_text("Loss: " + str(round(data.body["loss"], 3)))

            self.pos = 7
            for i, key in enumerate(sorted(data.body.keys())):
                if key not in ['loss', 'epoch']:
                    if i >= len(self.othertext):
                        self.othertext.append(self.ax.text(1, self.pos, "No key: "))
                    self.othertext[i].set_text(key[0].upper() + key[1:] + ": " + str(round(data.body[key], 3)))
                    self.pos -= 1

                if self.pos <= 0:
                    warnings.warn("Too many keys in loss metrics info.")
                    break

        elif data.mode == MessageMode.Test_End:
            self.epochtext.set_text("Training stats:")
            self.pos -= 1
            self.ax.text(1, self.pos, "Testing stats:")
            self.pos -= 1
            self.ax.text(1, self.pos, "Loss: " + str(round(data.body["loss"], 3)))
            self.pos -= 1
            for key in sorted(data.body.keys()):
                if key != "loss":
                    self.ax.text(1, self.pos, key[0].upper() + key[1:] + ": " + str(round(data.body[key], 3)))
                    self.pos -= 1


from MLDashboard.DashboardModules.ImageModule import ImageModule
from MLDashboard.MLCommunicationBackend import MessageMode

class TrainingSetSampleImages(ImageModule):
    def __init__(self, ax, config):
        """
        Shows a set of sample images from the training set.

        :param ax: matplotlib ax
        :param config: Config info with the keys width and height for the image and rows and cols for the plot
        """
        super().__init__(ax, config, "Training Set Sample Images", MessageMode.Train_Set_Sample)

    def initialRequests(self):
        return self.generateRequest()

    def update(self, data):
        if data.mode == MessageMode.Train_Set_Sample:
            images = self.createImages(data.body['x'])
            text = data.body['y']
            self.updateImageGrid(images, text)

        elif data.mode == MessageMode.Epoch_End:
            self.updateImageGrid()
            return self.generateRequest()



from MLDashboard.DashboardModules.Module import Module
from MLDashboard.MLCommunicationBackend import Message, MessageMode
from matplotlib.patches import Rectangle

class StatusModule(Module):
    """Module that shows current dashboard status info. Useful for monitoring performance."""
    def __init__(self, ax, config):
        super().__init__(ax, config, "Dashboard Status", noticks=True)
        self.ax.axis([0, 10, 0, 10])
        self.modetext = self.ax.text(1, 9, "Current Mode: ")
        self.autorendertext = self.ax.text(1, 8, "Autorendering: ")
        self.timertext = self.ax.text(1, 7, "Timer: ")

        self.rects = [] #holds comparative speed rects after first update call
        self.borderrects = []

    def update(self, data: Message):
        if data.mode == MessageMode.CustomData:
            self.modetext.set_text("Current Mode: " + str(data.body["currentmode"]))

            if data.body["currentmode"] == "Live Render":
                self.autorendertext.set_text("Autorendering: " + str(data.body["autorendering"]))
                self.timertext.set_text("Timer: " + str(data.body["timer"]))

            else:
                self.autorendertext.set_text("")
                self.timertext.set_text("")
                for rect in self.borderrects + self.rects:
                    rect.remove()

            leftx = 3
            rightx = 7
            topy = 5
            bottomy = 1

            rectw = (rightx - leftx) / data.body['width']
            recth = (topy - bottomy) / data.body['height']

            if len(self.rects) == 0:
                xpos = leftx
                ypos = topy
                counter = 0
                for i in range(0, len(data.body["modulestimer"])):
                    rect = Rectangle((xpos, ypos-recth), rectw, recth)
                    rectborder = Rectangle((xpos, ypos-recth), rectw, recth, fill=False, edgecolor='black')
                    counter += 1
                    xpos += rectw
                    if counter == data.body['width']:
                        counter = 0
                        ypos -= recth
                        xpos = leftx

                    self.rects.append(rect)
                    self.borderrects.append(rectborder)
                    self.ax.add_patch(rect)
                    self.ax.add_patch(rectborder)


            maxt = max(data.body["modulestimer"])
            for i, item in enumerate(data.body["modulestimer"]):
                if maxt != 0:
                    self.rects[i].set_width(item * rectw/maxt)
                else:
                    self.rects[i].set_width(0)


from MLDashboard.DashboardModules.Module import Module
from MLDashboard.MLCommunicationBackend import Message, MessageMode

class LossMetricsGraph(Module):
    """Graph showing current loss, accuracy, and other metrics."""
    def __init__(self, ax, config):
        super().__init__(ax, config, "Loss Metrics Graph")
        self.loss = []
        self.losscolor = ""
        self.metrics = {}
        self.metriccolors = {}

        self.ax.set_xlabel('Epoch')

    def update(self, data: Message):
        if data.mode == MessageMode.Epoch_End:
            self.loss.append(data.body['loss'])

            for key in data.body:
                if key not in ['loss', 'epoch']:
                    sublist = self.metrics.get(key, [])
                    sublist.append(data.body[key])
                    self.metrics[key] = sublist

            for key in self.metrics:
                if key not in self.metriccolors:
                    self.metriccolors[key] = self.ax.plot(range(1, len(self.metrics[key]) + 1),
                                                          self.metrics[key], label=key)[0].get_c()
                else:
                    self.ax.plot(range(len(self.metrics[key]) - 1, len(self.metrics[key]) + 1),
                                 self.metrics[key][-2:], color=self.metriccolors[key])

            if self.losscolor == "":
                self.losscolor = self.ax.plot(range(1, len(self.loss) +1), self.loss, label='loss')[0].get_c()
                self.ax.legend()
            else:
                self.ax.plot(range(len(self.loss) - 1, len(self.loss) + 1), self.loss[-2:], color=self.losscolor)

            ticks = list(range(0, len(self.loss), int((len(self.loss)-1)/10 + 1)))
            if len(ticks) == 1:
                ticks = [1]
            elif len(ticks) > 1 and ticks[1] == 1:
                for i in range(0, len(ticks)):
                    ticks[i] += 1
            else:
                ticks.append(ticks[-1] * 2 - ticks[-2])
            self.ax.set_xticks(ticks)
            self.ax.set_xlim(right=ticks[-1]+1)




import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2' #stops agressive error message printing
import tensorflow as tf
import MLDashboard.MLDashboardBackend as MLDashboardBackend
import MLDashboard.MLCallbacksBackend as MLCallbacksBackend
from MLDashboard.MLCommunicationBackend import Message, MessageMode
import time
from MLDashboard.Examples.InteractiveDashboardDemo import get_model

class myCustomCallback(MLCallbacksBackend.DashboardCallbacks):
    def __init__(self, updatelist, returnlist, model, x_train, y_train, x_test, y_test, labels, config):
        super().__init__(updatelist, returnlist, model, x_train, y_train, x_test, y_test, labels, config)

    def custom_on_test_begin(self, logs):
        print("We are beginning the evaluation step.")


def run(testmode = False):
    print("Starting custom callbacks demo...")
    print("Setting up dashboard...")

    #Create dashboard and return communication tools (this starts the process)
    dashboardjsonfile = os.path.dirname(__file__) + '/dashboarddemo.json'
    dashboardProcess, updatelist, returnlist = MLDashboardBackend.createDashboard(dashboardjsonfile,
                                                                                  openatend=not testmode)

    print("Loading data...")
    (x_train, y_train), (x_test, y_test) = tf.keras.datasets.mnist.load_data()

    print("Formatting data...")
    x_train = x_train.reshape(-1, 784).astype("float32") / 255.0
    x_test = x_test.reshape(-1, 784).astype("float32") / 255.0

    print("Sampling data...")
    # Limit the train data to 10000 samples
    x_train = x_train[:10000]
    y_train = y_train[:10000]
    # Limit test data to 1000 samples
    x_test = x_test[:1000]
    y_test = y_test[:1000]

    print("Creating model...")
    model = get_model()


    print("Creating custom callbacks...")
    #Callbacks require update and return list for communicating with dashboard
    #Model and datasets are useful for sending that data to certain modules
    config = MLCallbacksBackend.CallbackConfig()
    labels = ['zero', 'one', 'two', 'three', 'four', 'five', 'six', 'seven', 'eight', 'nine']
    callback = myCustomCallback(updatelist, returnlist, model, x_train, y_train, x_test, y_test, labels, config)

    print("Starting training...")
    trainingstarttime = time.time()
    model.fit(x_train, y_train, epochs=20, callbacks=[callback])
    print("Training finished in: ", round(time.time() - trainingstarttime, 3), " seconds.")

    print("Evaluating model...")
    model.evaluate(x_test, y_test, batch_size=128, callbacks=[callback])

    updatelist.append(Message(MessageMode.End, {}))
    print("Exiting cleanly...")
    dashboardProcess.join()
    print("Dashboard exited.")
    #This handles any extra data that the dashboard sent, such as save commands
    callback.HandleRemaingCommands()

if __name__ == '__main__':
    run()


from MLDashboard.DashboardModules.ImageModule import ImageModule
from MLDashboard.MLCommunicationBackend import MessageMode

class PredImages(ImageModule):
    def __init__(self, ax, config):
        """
        Shows a set of sample predictions from the prediction set.

        :param ax: matplotlib ax
        :param config: Config info with the keys width and height for the image and rows and cols for the plot
        """
        super().__init__(ax, config, "Sample Predictions", MessageMode.Pred_Sample)
        if "correctcolor" not in self.config:
            self.config['correctcolor'] = 'green'
        if "incorrectcolor" not in self.config:
            self.config['incorrectcolor'] = 'red'


    def update(self, data):
        if data.mode == MessageMode.Pred_Sample:
            images = self.createImages(data.body['x'])
            text = []
            color = []
            for i in range(0, len(images)):
                text.append(str(data.body['pred'][i]) + " : " + str(data.body['y'][i]))
                if str(data.body['pred'][i]) == str(data.body['y'][i]):
                    color.append(self.config['correctcolor'])
                else:
                    color.append(self.config['incorrectcolor'])

            self.updateImageGrid(images, text, color)

        elif data.mode == MessageMode.Epoch_End:
            self.updateImageGrid()
            return self.generateRequest()


import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2' #stops agressive error message printing
import tensorflow as tf
from tensorflow import keras
from MLDashboard.MLDashboardBackend import createDashboard
from MLDashboard.MLCallbacksBackend import DashboardCallbacks, CallbackConfig
from MLDashboard.MLCommunicationBackend import Message, MessageMode
import time

def get_model():
    model = keras.Sequential(
        [keras.layers.Dense(128, activation='relu'),
         keras.layers.Dense(10)]
    )

    model.compile(
        optimizer='adam',
        loss=keras.losses.SparseCategoricalCrossentropy(from_logits=True),
        metrics=["accuracy"],
    )

    return model

def run(testmode=False):
    print("Starting interactive dashboard demo...")
    print("Setting up dashboard...")

    #Create dashboard and return communication tools (this starts the process)
    dashboardjsonfile = os.path.dirname(__file__) + '/dashboarddemo.json'
    dashboardProcess, updatelist, returnlist = createDashboard(dashboardjsonfile, openatend=not testmode)

    print("Loading data...")
    (x_train, y_train), (x_test, y_test) = tf.keras.datasets.mnist.load_data()

    print("Formatting data...")
    x_train = x_train.reshape(-1, 784).astype("float32") / 255.0
    x_test = x_test.reshape(-1, 784).astype("float32") / 255.0

    print("Sampling data...")
    # Limit the train data to 10000 samples
    x_train = x_train[:10000]
    y_train = y_train[:10000]
    # Limit test data to 1000 samples
    x_test = x_test[:1000]
    y_test = y_test[:1000]

    print("Creating model...")
    model = get_model()


    print("Creating callbacks...")
    #Callbacks require update and return list for communicating with dashboard
    #Model and datasets are useful for sending that data to certain modules
    config = CallbackConfig()
    labels = list(range(0,10))
    #labels = ['zero', 'one', 'two', 'three', 'four', 'five', 'six', 'seven', 'eight', 'nine']
    callback = DashboardCallbacks(updatelist, returnlist, model, x_train, y_train, x_test, y_test, labels, config)

    print("Starting training...")
    trainingstarttime = time.time()
    model.fit(x_train, y_train, epochs=50, callbacks=[callback])
    print("Training finished in: ", round(time.time() - trainingstarttime, 3), " seconds.")

    print("Evaluating model...")
    model.evaluate(x_test, y_test, batch_size=128, callbacks=[callback])

    updatelist.append(Message(MessageMode.End, {}))
    print("Exiting cleanly...")
    dashboardProcess.join()
    print("Dashboard exited.")
    #This handles any extra data that the dashboard sent, such as save commands
    callback.HandleRemaingCommands()

if __name__ == '__main__':
    run()


#Module imports
from MLDashboard.DashboardModules.LossMetricsGraph import LossMetricsGraph
from MLDashboard.DashboardModules.LossMetricsNumerical import LossMetricsNumerical
from MLDashboard.DashboardModules.StatusModule import StatusModule
from MLDashboard.DashboardModules.ControlButtons import ControlButtons
from MLDashboard.DashboardModules.TrainingSetSampleImages import TrainingSetSampleImages
from MLDashboard.DashboardModules.PredImages import PredImages
from MLDashboard.DashboardModules.WrongPredImages import WrongPredImages
from MLDashboard.DashboardModules.EmptyModule import EmptyModule
from MLDashboard.DashboardModules.Module import Module

from MLDashboard.MLCommunicationBackend import Message, MessageMode
import matplotlib.pyplot as pyplot
from typing import List, Tuple
import multiprocessing
import json
import time

#region Dashboard
allModules = {'LossMetricsGraph': LossMetricsGraph,
              'LossMetricsNumerical': LossMetricsNumerical,
              'StatusModule': StatusModule,
              'ControlButtons': ControlButtons,
              'TrainingSetSampleImages': TrainingSetSampleImages,
              'PredImages': PredImages,
              'WrongPredImages': WrongPredImages,
              'EmptyModule': EmptyModule}

def dashboardProcess(configjson: dict, updatelist: list, returnlist: list, openatend):
    """Wrapper function to run dashboard in a seperate process. This should not be called manually."""
    print("Loading dashboard...")
    dashboard = Dashboard(configjson, updatelist, returnlist, openatend)
    print("Starting dashboard...")
    dashboard.runDashboardLoop()

def createDashboard(config='dashboard.json', waitforstart=True,
                    openatend=True) -> Tuple[multiprocessing.Process, List[Message], List[Message]]:
    """
    Creates a dashboard running in a seperate process.
    Returns the process, updatelist, and return list for communication
    :param config: The file to load the dashboard config from
    :param waitforstart: Should the main process halt while the dashboard starts
    :param openatend: Calls pyplot.show() at end of training
    """
    syncmanager = multiprocessing.Manager()
    updatelist: List[Message] = syncmanager.list()
    returnlist: List[Message] = syncmanager.list()

    with open(config) as f:
        configjson = json.load(f)
    process = multiprocessing.Process(target=dashboardProcess, args=(configjson, updatelist, returnlist, openatend,))
    process.start()

    if waitforstart:
        done = False
        rmindex = -1
        while not done:
            time.sleep(0.05)
            for index, item in enumerate(returnlist):
                if item.mode == MessageMode.Start:
                    rmindex = index
                    done = True
                    break

        returnlist.pop(rmindex)

    return process, updatelist, returnlist

def getModules(configjson):
    """Returns a list of module classes"""

    if "modules" not in configjson:
        raise Exception("Modules tag missing in config json.")
    modulelist = configjson["modules"]
    if len(modulelist) == 0:
        raise Exception("No modules found in config json.")

    sublistlen = len(modulelist[0])
    modules = []
    configs = []
    for sublist in modulelist:
        if len(sublist) != sublistlen:
            raise Exception("Modules do not form a grid.")
        for module, config in sublist:
            if module not in allModules:
                raise Exception("Module: " + str(module) + " not valid.")
            modules.append(allModules[module])
            configs.append(config)

    return modules, configs, sublistlen, len(modulelist)

def addRequests(reqs, outputlist):
    """Adds requests to output list if reqs is not none"""
    if reqs is not None:
        for item in reqs:
            outputlist.append(item)

class Dashboard:
    """
    Dashboard is a class that handles high level matplotlib interaction and sends data to sub modules.
    Dashbaords should be created with the createDashboard function.
    """
    def __init__(self, configjson: dict, updatelist: List[Message], returnlist: List[Message],
                 openatend:bool = False):
        self.configjson = configjson
        self.updatelist = updatelist
        self.returnlist = returnlist
        moduleclasslist, moduleconfiglist, self.width, self.height = getModules(configjson)
        self.modulelist: List[Module] = []


        self.fig = pyplot.figure()
        figManager = pyplot.get_current_fig_manager()
        figManager.window.state('zoomed')
        self.fig.suptitle("Tensorflow Dashboard")
        self.fig.canvas.manager.set_window_title("Tensorflow Dashboard")
        self.fig.set_tight_layout(True)
        for i, module in enumerate(moduleclasslist):
            ax = self.fig.add_subplot(self.height, self.width, i+1)
            c = moduleconfiglist[i]
            if 'config' in self.configjson:
                for key, value in self.configjson['config'].items():
                    if key not in c:
                        c[key] = value
            m = module(ax, c)
            addRequests(m.initialRequests(), self.returnlist)
            self.modulelist.append(m)

        # status metrics
        self.currentmode = "Live Render"  # are we rendering during training
        self.timer = 0  # time to do a full update loop
        self.modulestimer = [0.0] * len(self.modulelist) #how long does it take to render each module
        self.openatend = openatend

    def runDashboardLoop(self): #this function is designed to be run in a separate process
        """Continually updates modules"""
        done = False
        self.returnlist.append(Message(MessageMode.Start, {}))
        while not done:
            while len(self.updatelist) == 0:
                pyplot.draw()
                pyplot.pause(0.001) #only update when resting

            mostrecentupdate = self.updatelist.pop(0)
            if mostrecentupdate.mode == MessageMode.ForceUpdate:
                pyplot.draw()
                pyplot.pause(0.001)

            else:
                starttime = time.time()
                for i, module in enumerate(self.modulelist):
                    sTime = time.time()
                    self.updateModule(module, mostrecentupdate)
                    self.modulestimer[i] = round(time.time() - sTime, 3)
                self.timer = round(time.time() - starttime, 3)


            if mostrecentupdate.mode == MessageMode.End:
                done = True

        print("Dashboard exiting cleanly...")
        self.currentmode = 'Post Training View'
        for module in self.modulelist:
            self.updateModule(module, Message(MessageMode.End, {}))

        if self.openatend:
            pyplot.show()
        for module in self.modulelist:
            req = module.update(Message(MessageMode.End, {}))
            if req is not None:
                for item in req:
                    self.returnlist.append(item)

    def updateModule(self, module, mostrecentupdate: Message):
        if type(module) == StatusModule:
            reqs = module.update(Message(MessageMode.CustomData, {'autorendering': len(self.updatelist) <= 1,
                                                                  'currentmode': self.currentmode,
                                                                  'timer': self.timer,
                                                                  'modulestimer': self.modulestimer,
                                                                  'width': self.width,
                                                                  'height': self.height}))
        else:
            reqs = module.update(mostrecentupdate)

        addRequests(reqs, self.returnlist)

#endregion